from google.cloud.logging.handlers import CloudLoggingHandler


class StackdriverHandler(CloudLoggingHandler):
    """Logging handler for Google Cloud Platform Stackdriver Logging.
    This handler is inherited from google.cloud.logging.handlers.CloudLoggingHandler
    Stackdriver has a limit for the log message to be 256KB
    This handler truncates the message if it is over 200KB

    See Also:
    https://github.com/googleapis/google-cloud-python/blob/master/logging/google/cloud/logging/handlers/handlers.py
    
    Args:
        CloudLoggingHandler ([type]): [description]
    """
    def emit(self, record):
        limit = 200 * 1024
        # Convert the log message to string
        record.msg = str(record.msg)
        if len(record.msg) > limit:
            record.msg = record.msg[:limit]
        return super().emit(record)



import os
import datetime
import logging
from io import FileIO
from abc import ABC
from .base import StorageObject, StoragePrefixBase, StorageIOSeekable
logger = logging.getLogger(__name__)


class BucketStorageObject(StorageObject):
    """Represents a cloud storage object associated with a bucket.
    This object may not correspond to an actual object in the bucket, e.g. a folder in Google or S3 bucket.
    """
    # Caches clients for each scheme
    cache_dict = dict()
    # Expiration time for each client
    cache_expire = dict()

    # The number of seconds before the client expires.
    CACHE_EXPIRE_SEC = 1200

    def __init__(self, uri):
        StorageObject.__init__(self, uri)
        self._client = None
        self._bucket = None
        self._blob = None

    @classmethod
    def get_cached(cls, obj_id, init_method):
        """Gets an unexpired object by obj_id from cache, creates one using init_method() if needed.
        """
        cached_obj = cls.cache_dict.get(obj_id)
        now = datetime.datetime.now()
        if cached_obj:
            client_expire = cls.cache_expire.get(obj_id)
            # Use the cached client if it is not expired.
            if client_expire and client_expire > now:
                return cached_obj
        obj = init_method()
        cls.cache_dict[obj_id] = obj
        cls.cache_expire[obj_id] = now + datetime.timedelta(seconds=cls.CACHE_EXPIRE_SEC)
        return obj

    def get_client(self):
        obj_id = self.scheme
        return self.get_cached(obj_id, self.init_client)

    def get_bucket(self):
        obj_id = "%s://%s" % (self.scheme, self.bucket_name)
        return self.get_cached(obj_id, self.init_bucket)

    @property
    def bucket_name(self):
        """The name of the Cloud Storage bucket as a string."""
        return self.hostname

    @property
    def client(self):
        if not self._client:
            self._client = self.get_client()
        return self._client

    @property
    def bucket(self):
        if not self._bucket:
            self._bucket = self.get_bucket()
        return self._bucket

    def is_file(self):
        """Determine if the object is a file.
        This will return False if the object does not exist or the object is a folder.
        """
        if self.path.endswith("/"):
            return False
        if not self.exists():
            return False
        return True

    def init_client(self):
        raise NotImplementedError()

    def init_bucket(self):
        raise NotImplementedError()

    def exists(self):
        raise NotImplementedError()


class CloudStoragePrefix(StoragePrefixBase, ABC):
    def blobs(self, delimiter=""):
        """All blobs with the same prefix as this object
        The type of blobs depends on the actual implementation of the blobs() method.
        The delimiter causes a list operation to roll up all the keys that share a common prefix into a single result.
        See Also: https://docs.aws.amazon.com/AmazonS3/latest/dev/ListingKeysHierarchy.html
        """
        raise NotImplementedError()


class CloudStorageIO(StorageIOSeekable):
    def __init__(self, uri):
        """
        """
        StorageIOSeekable.__init__(self, uri)

        # Path of the temp local file
        self.temp_path = None

        # Stores the temp local FileIO object
        self.__file_io = None

    @property
    def size(self):
        if self.__file_io:
            return os.fstat(self.__file_io.fileno).st_size
        return self.get_size()

    def seek(self, pos, whence=0):
        if self.__file_io:
            self._offset = self.__file_io.seek(pos, whence)
            return self._offset
        return self._seek(pos, whence)

    def tell(self):
        if self.__file_io:
            self._offset = self.__file_io.tell()
        return self._offset

    def local(self):
        """Creates a local copy of the file.
        """
        if not self.__file_io:
            file_obj = self.create_temp_file()
            # Download file if appending or updating
            if self.exists() and ('a' in self.mode or '+' in self.mode):
                self.download(file_obj)
            # Close the temp file and open it with FileIO
            file_obj.close()
            mode = "".join([c for c in self.mode if c in "rw+ax"])
            self.__file_io = FileIO(file_obj.name, mode)
            self.temp_path = file_obj.name
        return self

    def read(self, size=None):
        """Reads the file from the Google Cloud bucket to memory

        Returns: Bytes containing the contents of the file.
        """
        start = self.tell()
        if self.__file_io:
            self.__file_io.seek(start)
            b = self.__file_io.read(size)
        else:
            if not self.exists():
                raise FileNotFoundError("File %s does not exists." % self.uri)
            file_size = self.size
            # TODO: size unknown?
            if not file_size:
                return b""
            if start >= file_size:
                return b""
            end = file_size - 1
            if size:
                end = start + size - 1
            if end > file_size - 1:
                end = file_size - 1
            # logger.debug("Reading from %s to %s" % (start, end))
            b = self.read_bytes(start, end)
        self._offset += len(b)
        return b

    def write(self, b):
        """Writes data into the file.

        Args:
            b: Bytes data

        Returns: The number of bytes written into the file.

        """
        if self.closed:
            raise ValueError("write to closed file %s" % self.uri)
        # Create a temp local file
        self.local()
        # Write data from buffer to file
        self.__file_io.seek(self.tell())
        size = self.__file_io.write(b)
        self._offset += size
        return size

    def __rm_temp(self):
        if self.temp_path and os.path.exists(self.temp_path):
            os.unlink(self.temp_path)
        logger.debug("Deleted temp file %s of %s" % (self.temp_path, self.uri))
        self.temp_path = None
        return

    def open(self, mode='r', *args, **kwargs):
        """Opens the file for writing
        """
        if not self._closed:
            self.close()
        super().open(mode)
        self._closed = False
        # Reset offset position when open
        self.seek(0)
        if 'a' in self.mode:
            # Move to the end of the file if open in appending mode.
            self.seek(0, 2)
        elif 'w' in self.mode:
            # Create empty local file
            self.local()
        return self

    def close(self):
        """Flush and close the file.
        This method has no effect if the file is already closed.
        """

        if self._closed:
            return

        if self.__file_io:
            if not self.__file_io.closed:
                self.__file_io.close()
            self.__file_io = None

        if self.temp_path:
            logger.debug("Uploading file to %s" % self.uri)
            with open(self.temp_path, 'rb') as f:
                self.upload(f)
            # Remove __temp_file if it exists.
            self.__rm_temp()
            # Set _closed attribute
            self._closed = True

    @property
    def updated_time(self):
        raise NotImplementedError()

    def exists(self):
        raise NotImplementedError()

    def get_size(self):
        raise NotImplementedError()

    def delete(self):
        raise NotImplementedError()

    def upload(self, from_file_obj):
        raise NotImplementedError()

    def download(self, to_file_obj):
        """Downloads the data to a file object
        Caution: This method does not call flush()
        """
        raise NotImplementedError()

    def read_bytes(self, start, end):
        """Reads bytes from position start to position end, inclusive
        """
        raise NotImplementedError()



"""Contains tests for the tasks module.
"""
import time
import logging

import os
import sys
aries_parent = os.path.join(os.path.dirname(__file__), "..", "..")
if aries_parent not in sys.path:
    sys.path.append(aries_parent)
from Aries.test import AriesTest
from Aries import tasks

logger = logging.getLogger(__name__)


class TestShellCommand(AriesTest):
    def test_run_shell_command(self):
        cmd = tasks.ShellCommand("ls -a %s" % os.path.dirname(__file__))
        cmd.run()
        self.assertIn("..\n", cmd.std_out)
        self.assertIn(os.path.basename(__file__), cmd.std_out)


class TestRunRetry(AriesTest):
    tries = 0

    @staticmethod
    def func_to_retry():
        if TestRunRetry.tries < 2:
            TestRunRetry.tries += 1
            print("Test Run Retry: %s" % TestRunRetry.tries)
            raise ValueError("Try again later")
        return TestRunRetry.tries

    def setUp(self):
        TestRunRetry.tries = 0

    def test_run_and_retry_success(self):
        """Tests the run and retry of FunctionTask
        """
        # Running the function for the first time will raise a ValueError
        with self.assertRaises(ValueError):
            self.func_to_retry()
        # Retry the runing the function until there is no error.
        task = tasks.FunctionTask(self.func_to_retry)
        count = task.run_and_retry(3)
        self.assertEqual(count, 2)

    def test_run_and_retry_fail(self):
        with self.assertRaises(ValueError):
            task = tasks.FunctionTask(self.func_to_retry)
            task.run_and_retry(2)

class TestFunctionTask(AriesTest):

    @staticmethod
    def func_with_delay(name, delay):
        """A function with outputs and logs.
        The function will sleep for the seconds specified in "delay" between
            sending two sets of outputs and logs.

        Args:
            name (str): An identifier to be included in the outputs and log messages.
            delay (int): The number of seconds between sending two sets of outputs and logs.

        Returns: The "name" as a string.

        """
        print("%s Function Started" % name)
        TestFunctionTask.output_logs(name, "BEFORE DELAY")
        print("%s Std Error Test" % name, file=sys.stderr)
        for i in range(delay):
            time.sleep(1)
            print("%s waited %s second" % (name, i + 1))
        TestFunctionTask.output_logs(name, "AFTER DELAY")
        print("%s Function Ended" % name)
        return name

    @staticmethod
    def func_with_exception(name, delay):
        """A function with outputs, logs and an exception.
        The function will sleep for the seconds specified in "delay" between
            sending outputs/logs and raising an exception.

        Args:
            name (str): An identifier to be included in the outputs and log messages.
            delay (int): The number of seconds between sending outputs/logs and raising an exception.

        Returns: This function does not return anything.

        """
        print("%s Function Started" % name)
        TestFunctionTask.output_logs(name)
        print("%s Std Error Test" % name, file=sys.stderr)
        for i in range(delay):
            time.sleep(1)
            print("%s waited %s second" % (name, i + 1))
        raise Exception("%s Exception" % name)

    @staticmethod
    def output_logs(prefix, suffix=""):
        """Outputs five levels of logs. Each log message will have the specific prefix and suffix.

        Args:
            prefix (str): Prefix for the log messages.
            suffix (str): Suffix for the log messages.

        Returns: This function does not return anything.

        """
        logger.debug("%s DEBUG LOG %s" % (prefix, suffix))
        logger.info("%s INFO LOG %s" % (prefix, suffix))
        logger.warning("%s WARNING LOG %s" % (prefix, suffix))
        logger.error("%s ERROR LOG %s" % (prefix, suffix))
        logger.critical("%s CRITICAL LOG %s" % (prefix, suffix))

    @staticmethod
    def expected_logs(prefix, suffix=""):
        """Gets the expected log messages output by output_logs() as a list.

        Args:
            prefix (str): Prefix for the log messages.
            suffix (str): Suffix for the log messages.

        Returns: A list of strings.

        """
        return [
                   "%s %s LOG %s" % (prefix, level, suffix)
                   for level in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
               ]

    def assert_task_output(self, output_string, expected_list):
        """Checks if a list of strings is a subset (sub-list) of  another list of strings.

        Args:
            output_string: A list of string expected to contain strings in another list.
            expected_list: The smaller list of string to be contained in another list of strings.

        This function will raise exception if not all strings in the expected_list are found in the output_string.

        Returns: This function does not return anything.

        """
        output_list = output_string.strip("\n").split("\n")
        for msg in expected_list:
            for out in output_list:
                if msg in out:
                    break
            else:
                self.fail("Output not found: %s" % msg)

    def test_run_tasks_normal(self):
        """Tests running two functions and capture their outputs independently.

        In this test, t2 output logs and messages while t1 is running,
        the outputs from t2 should not go into t1's output.

        Also, t1 and t2 should not capture output any from the main thread.
        """
        print()
        # t1 will be executed first with a longer delay.
        func_name_1 = "test1"
        t1 = tasks.FunctionTask(self.func_with_delay, func_name_1, 2)
        t1.run_async()
        time.sleep(0.5)
        # t2 will be executed with a shorter delay.
        func_name_2 = "test2"
        t2 = tasks.FunctionTask(self.func_with_delay, func_name_2, 1)
        t2.run_async()
        # Output messages on the main thread.
        # These messages should not go into the outputs of t1 or t2
        print("Message to STDERR", file=sys.stderr)
        logger.info("Main Thread Log")
        print("WAITING for t1")
        # Wait for the functions to terminate.
        # t2 will terminate before t1.
        t1.join()
        print("WAITING for t2")
        t2.join()

        # Print out the captured outputs, for debug purpose.
        t1.print_outputs()
        t2.print_outputs()

        # Assert return values
        self.assertEqual(t1.returns, func_name_1, "Incorrect return value.")
        self.assertEqual(t2.returns, func_name_2, "Incorrect return value.")

        # Assert standard outputs
        expected_outputs_t1 = [
            "%s Function Started" % func_name_1,
            "%s Function Ended" % func_name_1,
        ]
        expected_outputs_t2 = [
            "%s Function Started" % func_name_2,
            "%s Function Ended" % func_name_2,
        ]
        self.assert_task_output(t1.std_out, expected_outputs_t1)
        self.assert_task_output(t2.std_out, expected_outputs_t2)

        # Assert standard error
        self.assert_task_output(t1.std_err, ["%s Std Error Test" % func_name_1])
        self.assert_task_output(t2.std_err, ["%s Std Error Test" % func_name_2])

        # Assert logs
        # Expected logs
        expected_msgs_t1 = self.expected_logs(func_name_1, "BEFORE DELAY") + \
            self.expected_logs(func_name_1, "AFTER DELAY")
        expected_msgs_t2 = self.expected_logs(func_name_2, "BEFORE DELAY") + \
            self.expected_logs(func_name_2, "AFTER DELAY")
        # Check if the expected logs are in the actual logs
        self.assert_task_output(t1.log_out, expected_msgs_t1)
        self.assert_task_output(t2.log_out, expected_msgs_t2)
        # Check the number of logs messages
        self.assertEqual(len(expected_msgs_t1), len(t1.log_list))
        self.assertEqual(len(expected_msgs_t2), len(t2.log_list))

    def test_run_tasks_with_exception(self):
        """Tests running two functions with exceptions and capture their outputs independently.
        """
        print()
        # t1 will be executed first with a longer delay.
        func_name_1 = "test1"
        t1 = tasks.FunctionTask(self.func_with_exception, func_name_1, 2)
        t1.run_async()
        # t2 will be executed with a shorter delay.
        func_name_2 = "test2"
        t2 = tasks.FunctionTask(self.func_with_exception, func_name_2, 1)
        t2.run_async()
        # Print a message from the main thread.
        # This message should not go into the outputs of t1 or t2
        print("Message to STDERR", file=sys.stderr)
        logger.info("Main Thread Log")
        print("WAITING for t1")
        # Wait for the functions to terminate.
        # t2 will terminate before t1.
        t1.join()
        print("WAITING for t2")
        t2.join()

        # Print out the captured outputs.
        t1.print_outputs()
        t2.print_outputs()

        # Assert return values
        self.assertIsNone(t1.returns, "Return value should be None.")
        self.assertIsNone(t2.returns, "Return value should be None.")

        # Assert standard outputs
        expected_outputs_t1 = [
            "%s Function Started" % func_name_1,
        ]
        expected_outputs_t2 = [
            "%s Function Started" % func_name_2,
        ]
        self.assert_task_output(t1.std_out, expected_outputs_t1)
        self.assert_task_output(t2.std_out, expected_outputs_t2)

        # Assert standard error
        self.assert_task_output(t1.std_err, ["%s Std Error Test" % func_name_1])
        self.assert_task_output(t2.std_err, ["%s Std Error Test" % func_name_2])

        # Assert logs
        expected_msgs_t1 = self.expected_logs(func_name_1)
        expected_msgs_t2 = self.expected_logs(func_name_2)
        self.assert_task_output(t1.log_out, expected_msgs_t1)
        self.assert_task_output(t2.log_out, expected_msgs_t2)

        # Assert Exception
        self.assertIn("test1 Exception", t1.exc_out)
        self.assertIn("test2 Exception", t2.exc_out)



import binascii
import gzip
import os
import json
import tempfile
import logging
import re
from urllib.parse import urljoin, urlparse
from shutil import copyfile, copyfileobj
from .strings import FileName

logger = logging.getLogger(__name__)


class File:
    """Provides shortcuts for handling files.
    """

    __signatures = None
    __sign_size = None

    @staticmethod
    def load_json(file_path, default=None):
        """Loads data from a JSON file to a Python dictionary

        Args:
            file_path: file path of a json file.
            default: default value to be returned if the file does not exist.
                If default is None and file is not found , an empty dictionary will be returned.

        Returns: A python dictionary containing data from the json file.
        """
        if os.path.exists(file_path):
            with open(file_path) as f:
                data = json.load(f)
        else:
            logger.info("File %s Not Found." % file_path)
            if default:
                data = default
            else:
                data = {}
        return data

    @staticmethod
    def load_signatures(json_path=None):
        """Loads the file signature dictionary.
        
        See Also: helpers.file_signatures.py
        
        Args:
            json_path (str, optional): Path of the JSON file containing the signature dictionary. 
            If json_path is None, "./assets/file_signatures.json" will be used.
        
        Returns:
            dict: Signature dictionary
        """
        if json_path is None:
            json_path = os.path.join(os.path.dirname(__file__), "assets", "file_signatures.json")
        with open(json_path, "r") as f:
            signatures = json.load(f)
        return signatures

    def __init__(self, file_path):
        self.file_path = file_path

    @property
    def signatures(self):
        """Signature dictionary

        Each key is the number of offset bytes.
        Each value is a dictionary with signatures(hex) as keys and file mine as values.
        
        Returns:
            dict: Signature dictionary
        """
        if self.__signatures is None:
            self.__signatures = self.load_signatures()
            self.__sign_size = {}
            for offset, v in self.__signatures.items():
                signs = v.keys()
                max_size = 0
                for sign in signs:
                    size = len(sign)
                    max_size = size if size > max_size else max_size
                self.__sign_size[offset] = max_size
        return self.__signatures

    def hex(self, size, offset=0):
        """Reads bytes from the file as hexadecimal string
        
        Args:
            size (int): The number of bytes to read.
            offset (int, optional): The offset to start reading. Defaults to 0.
        
        Returns:
            str: Return the hexadecimal representation of the binary data. 
            Every byte of data is converted into the corresponding 2-digit hex representation. 
            The resulting string is therefore twice of size.
        """
        with open(self.file_path, 'rb') as f:
            if offset:
                f.read(offset)
            return binascii.hexlify(f.read(size))

    def file_type(self):
        """Tries to identify the file type
        
        Returns:
            str: file mime type if identified, otherwise None.
        """
        signatures = self.signatures
        for offset, max_size in self.__sign_size.items():
            signs = sorted(signatures.get(offset).keys(), reverse=True)
            hex_value = self.hex(max_size, offset=int(offset)).decode().upper()
            for sign in signs:
                if hex_value.startswith(sign.upper()):
                    return signatures.get(offset).get(sign)
        return None

    def unzip(self, to_path=None):
        """Un-zips a .gz file within the same directory.

        Args:
            to_path (str): the output file path to store the unzipped file.

        Returns:
            The output filename with full path.
        """
        if to_path is None:
            output_file = self.file_path[:-3]
        else:
            output_file = to_path
        with gzip.open(self.file_path, 'rb') as gzip_file:
            with open(output_file, "wb") as unzipped_file:
                logger.debug("Unzipping %s to %s ..." % (self.file_path, output_file))
                block_size = 1 << 20
                while True:
                    block = gzip_file.read(block_size)
                    if not block:
                        break
                    unzipped_file.write(block)
        return output_file

    def gzip(self, to_path=None):
        if to_path is None:
            to_path = self.file_path + ".gz"
        with open(self.file_path, 'rb') as f_in:
            with gzip.open(to_path, 'wb') as f_out:
                copyfileobj(f_in, f_out)
        return to_path


class Markdown:
    """Represents a Markdown file.
    """

    # A list of regex patterns for links,
    # If there is a match, group(1) must be the url
    link_patterns = [
        # Inline link
        r"\[.+\]\((\S+)( [\"\'\(].+[\"\'\)])?\)",
        # HTML link
        r"<a .*?[\n]?.*?href=[\"\'](.+?)[\"\'].*?[\n]?.*?>.*?[\n]?.*?</a>",
        # Reference link with title in a new line
        r"\[.+\]:[ \t]*(\S+)\s*([\"\'\(].+[\"\'\)])",
        # Reference link
        r"\[.+\]:[ \t]*(\S+)[ \t]*([\"\'\(].+[\"\'\)])?",
    ]

    def __init__(self, file_path=None):
        """Initialize a Markdown object
        
        Args:
            file_path (str): File path of a Markdown(.md) file.
        """
        self.file_path = file_path
        if not self.file_path:
            self.text = None
            return
        with open(self.file_path) as f:
            self.text = f.read()

    @staticmethod
    def from_text(s):
        md = Markdown()
        md.text = s
        return md

    @property
    def title(self):
        """Gets the first title of the Markdown file.
        """
        lines = self.text.split("\n")
        for line in lines:
            if line.startswith("#"):
                return line.strip("#").strip("\n").strip()
        return ""

    def find_links(self):
        """Finds all links in the markdown.

        Returns: A list of 2-tuples. In each tuple,
            The first element is the string matching a link pattern
            The second element is the link URL

        """
        patterns = "|".join(["(%s)" % pattern for pattern in self.link_patterns])
        match_iter = re.finditer(patterns, self.text)
        matches = list(match_iter)
        match_list = []
        for match in matches:
            groups = [element for element in match.groups() if element is not None]
            logger.debug(groups)
            if len(groups) > 1:
                match_list.append((groups[0], groups[1]))
            else:
                match_list.append((groups[0], None))
        return match_list

    def make_links_absolute(self, base_url):
        """Converts all links in the file to absolute links.

        Returns:

        """
        replace_dict = {}
        links = self.find_links()
        for link in links:
            text = link[0]
            url = link[1]
            result = urlparse(url)
            if not result.scheme:
                abs_url = urljoin(base_url, url)
                replace_dict[text] = text.replace(url, abs_url)
        for key, val in replace_dict.items():
            self.text = re.sub(re.escape(key), val, self.text, 1)
        return self.text


class TemporaryFile:
    # TODO: Inherit from NamedTemporaryFile
    """Represents a temporary file.

    Examples:
        with TemporaryFile(template) as temp_file:
            ...

    If "template" is None, a new tempfile.NamedTemporaryFile() will be created.

    This will create a temp_file by copying an existing file (template).
    The temp_file will be deleted when exiting the "with"

    """
    def __init__(self, template=None):
        """Initialize with the file path of a template.

        Args:
            template: A template for creating the temporary file.
                The temp file will be a copy of the template.
                A empty NamedTemporaryFile() will be created if template is None.

        """
        self.template = template
        self.temp_file = "None"
    
    def filename(self):
        return os.path.basename(self.temp_file)

    def __enter__(self):
        """Creates a temp file by copying the template, if any.

        Returns: The full path of the temp file.

        """
        if self.template:
            filename = os.path.basename(self.template)
            filename = FileName(filename).append_random_letters(8).to_string()

            temp_folder = tempfile.gettempdir()
            self.temp_file = os.path.join(temp_folder, filename)
            copyfile(self.template, self.temp_file)
        else:
            f = tempfile.NamedTemporaryFile(delete=False)
            f.close()
            self.temp_file = f.name
        return self.temp_file

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Removes the temp_file.
        """
        if os.path.exists(self.temp_file):
            os.remove(self.temp_file)



import os
import sys
import logging
from unittest import TestCase
aries_parent = os.path.join(os.path.dirname(__file__), "..", "..")
if aries_parent not in sys.path:
    sys.path.append(aries_parent)
from Aries.test import AriesTest
from Aries.web import WebAPI
from Aries.outputs import CaptureOutput, StreamHandler, LoggingConfig, PackageLogFilter


class TestLogging(AriesTest):
    def test_log_formatter(self):
        """Tests log formatter for list and dict.
        """
        logger = logging.getLogger(__name__)
        with CaptureOutput(suppress_exception=True) as out:
            logger.debug(["A", "B"])
            logger.debug(dict(a=1, b=2))
            logger.debug("ABC")
        list_log = out.logs[0].split("\n", 1)[1].strip("\n")
        dict_log = out.logs[1].split("\n", 1)[1].strip("\n")
        self.assertEqual(list_log, '[\n    "A",\n    "B"\n]')
        self.assertEqual(dict_log, '{\n    "a": 1,\n    "b": 2\n}')
        self.assertNotIn("\n", out.logs[2])

    def test_capturing_outputs(self):
        """Tests capturing outputs using the CaptureOutput class
        """
        logger = logging.getLogger(__name__)
        # Test setting the level to INFO and there should be no DEBUG log
        with CaptureOutput(log_level=logging.INFO) as out:
            logger.debug("Test Debug")
            logger.info(b"Test Info")
            print("Test Print")
        # There should be only one log since the level was set to INFO
        self.assertEqual(len(out.logs), 1, out.logs)
        self.assertTrue(out.logs[0].endswith("Test Info"), out.logs[0])
        self.assertEqual(out.std_out, "Test Print\n")


class TestConfigLogging(TestCase):
    """Tests here only checks whether the program executed without error.
    They do not check whether the logging outputs are correct.
    
    """
    def test_context_manager(self):
        """Tests config logging with context manager
        """
        logger = logging.getLogger(__name__)
        with LoggingConfig():
            logger.debug("Test Config Debug")
            logger.info("Test Config Info")
        logger.debug("Logging should not show up.")

    def test_enable_disable(self):
        logger = logging.getLogger(__name__)
        config = LoggingConfig(formatter='%(message)s').enable()
        logger.debug("Logging Enabled")
        config.disable()
        logger.debug("Logging should not show up.")
    
    def test_decorator(self):
        """Tests config logging with decorator.
        """
        @LoggingConfig.decorate
        def test_func(msg):
            logger = logging.getLogger(__name__)
            logger.debug("Test Decorator: %s" % msg)
        
        logger = logging.getLogger(__name__)
        logger.debug("Logging should not show up.")
        test_func("debug")

    def test_log_filter(self):
        """Test config logging with filter
        """
        package_filter = PackageLogFilter(packages="tests")
        with LoggingConfig(filters=[package_filter]):
            logger = logging.getLogger(__name__)
            logger.debug("tests package Debug")
            WebAPI("https://www.google.com").get("")
            logger.info("tests package Info")
        logger.debug("Logging should not show up.")



import logging
import boto3
from botocore.exceptions import ClientError
from .base import StorageFolderBase
from .cloud import BucketStorageObject, CloudStoragePrefix, CloudStorageIO
logger = logging.getLogger(__name__)


class S3Object(BucketStorageObject):
    """
    See Also: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html
    """

    @property
    def key(self):
        return self.prefix

    @property
    def blob(self):
        """Gets or initialize a S3 Storage Object.

        Returns: An S3 storage object.

        This does not check whether the object exists.
        Use blob.exists() to determine whether or not the blob exists.

        """
        s3 = boto3.resource('s3')
        # logger.debug("Getting blob: %s" % self.uri)
        return s3.Object(self.bucket_name, self.prefix)

    def init_client(self):
        return boto3.client('s3')

    def init_bucket(self):
        s3 = boto3.resource('s3')
        logger.debug("Getting bucket %s" % self.bucket_name)
        return s3.Bucket(self.bucket_name)

    def exists(self):
        """

        Returns:

        See Also:
            https://boto3.amazonaws.com/v1/documentation/api/latest/guide/migrations3.html#accessing-a-bucket
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.head_object
        """
        try:
            self.client.head_object(Bucket=self.bucket_name, Key=self.prefix)
            return True
        except ClientError as e:
            error_code = e.response['Error']['Code']
            if error_code == '404':
                return False
            raise e

    def create(self):
        return self.client.put_object(Bucket=self.bucket_name, Key=self.prefix)

    def delete(self):
        return self.blob.delete()


class S3Prefix(CloudStoragePrefix, S3Object):
    @property
    def uri_list(self):
        keys = []
        response = self.client.list_objects(Bucket=self.bucket_name, Prefix=self.prefix, Delimiter='')
        contents = response.get("Contents", [])
        keys.extend([element.get("Key") for element in contents])
        return [
            "s3://%s/%s" % (self.bucket_name, key)
            for key in keys
            if not key.endswith("/")
        ]

    def blobs(self, delimiter=""):
        return list(self.bucket.objects.filter(Prefix=self.prefix, Delimiter=delimiter))

    def exists(self):
        response = self.client.list_objects(Bucket=self.bucket_name, Prefix=self.prefix)
        contents = response.get("Contents", []) if response else None
        if contents:
            return True
        return False

    def delete(self):
        return self.bucket.objects.filter(Prefix=self.prefix).delete()


class S3Folder(S3Prefix, StorageFolderBase):
    def __init__(self, uri):
        """Initializes a Google Cloud Storage Directory.

        Args:
            uri: The path of the object, e.g. "gs://bucket_name/path/to/dir/".

        """
        # super() will call the __init__() of StorageObject, StorageFolder and GSObject
        S3Object.__init__(self, uri)
        StorageFolderBase.__init__(self, uri)

        # Make sure prefix ends with "/", otherwise it is not a "folder"
        if self.prefix and not self.prefix.endswith("/"):
            self.prefix += "/"

    @property
    def folder_paths(self):
        """Folders(Directories) in the directory.
        """
        return self.__folders_paths()

    def __folders_paths(self):
        # TODO: Get the next page
        folders = []
        response = self.client.list_objects(Bucket=self.bucket_name, Prefix=self.prefix, Delimiter='/')
        prefixes = response.get("CommonPrefixes", [])
        folders.extend([p.get('Prefix') for p in prefixes if p.get('Prefix')])
        return [
            "s3://%s/%s" % (self.bucket_name, p)
            for p in folders
        ]

    @property
    def file_paths(self):
        """Files in the directory
        """
        paths = self.__file_paths()
        return paths

    def __file_paths(self):
        keys = []
        response = self.client.list_objects(Bucket=self.bucket_name, Prefix=self.prefix, Delimiter='/')
        contents = response.get("Contents", [])
        keys.extend([element.get("Key") for element in contents])
        return [
            "s3://%s/%s" % (self.bucket_name, key)
            for key in keys
            if not key.endswith("/")
        ]


class S3File(S3Object,CloudStorageIO):
    def __init__(self, uri):
        # file_io will be initialized by open()
        # self.file_io = None
        S3Object.__init__(self, uri)
        CloudStorageIO.__init__(self, uri)

    @property
    def updated_time(self):
        return self.blob.last_modified

    @property
    def md5_hex(self):
        e_tag = str(self.blob.e_tag)
        if len(e_tag) > 2:
            return e_tag.strip("\"")
        return None

    def get_size(self):
        return self.blob.content_length

    def upload(self, from_file_obj):
        self.blob.upload_fileobj(from_file_obj)

    def download(self, to_file_obj):
        self.blob.download_fileobj(to_file_obj)
        return to_file_obj

    def read_bytes(self, start, end):
        response = self.blob.get(Range="bytes=%s-%s" % (start, end))
        content = response.get("Body")
        if content:
            data = content.read()
            return data
        return ""

    def copy(self, to):
        dest = S3File(to)
        logger.debug("Creating file %s" % to)
        return dest.blob.copy(dict(Bucket=self.bucket_name, Key=self.prefix))
